"""This module implements the Aug class."""
import uuid

from rook.logger import logger
from rook.processor.error import Error
from rook.processor.namespaces.container_namespace import ContainerNamespace
from rook.processor.namespaces.stack_namespace import StackNamespace
from rook.processor.namespaces.python_utils_namespace import PythonUtilsNamespace
from rook.processor.namespaces.trace_namespace import TraceNamespace
from rook.processor.namespaces.process_state_namespace import ProcessStateNamespace
from rook.user_warnings import UserWarnings

from rook.exceptions import RookRuleMaxExecutionTimeReached, RookRuleRateLimited
from rook.augs.aug_rate_limiter import AugRateLimiter
from rook.config import RateLimiter
import rook.utils as utils


class Aug(object):
    """The Aug class is the skeleton that holds together all the components that define a modification to the application.

    This class brings together the following elements:
    - location - specifies when to run the modification.
    - extractor - specifies attributes to extract from the application's state before evaluating the modification.
    - condition - specifies an optional filter as to when to run the modification.
    - action - specifies the modification to preform.
    """

    def __init__(self, aug_id, extractor, condition, action, max_aug_execution_time_ns=0, rate_limits_ns=None):
        """Build an Aug object from the individual elements."""
        self.aug_id = aug_id
        self._extractor = extractor
        self._condition = condition
        self._action = action
        self._max_aug_time = max_aug_execution_time_ns
        self._enabled = True

        if rate_limits_ns:
            self._rate_limiter = AugRateLimiter(rate_limits_ns[0], rate_limits_ns[1])
        else:
            self._rate_limiter = AugRateLimiter(None, None)

    def execute(self, frame_namespace, extracted, output):
        """Called by the trigger service to run the extractor, condition and action."""
        if not self._enabled:
            return

        if output.is_user_messages_queue_full():
            logger.warn("Skipping aug ({}) execution because the queue is full".format(self.aug_id))
            return

        now_ns = utils.get_most_accurate_time_stamp_nano_seconds()
        limit_key = self._rate_limiter.allow(now_ns)

        try:
            if self._extractor:
                self._extractor.execute(frame_namespace, extracted)

            store = ContainerNamespace({})

            namespace = ContainerNamespace({
                'frame': frame_namespace,
                'stack': StackNamespace(frame_namespace),
                'extracted': ContainerNamespace(extracted),
                'store': store,
                'temp': ContainerNamespace({}),
                'python': PythonUtilsNamespace(),
                'utils': PythonUtilsNamespace(),
                'trace': TraceNamespace(),
                'state': ProcessStateNamespace(),
            })

            # TODO - cleanup rate limit
            if not self._condition or self._condition.evaluate(namespace, extracted):
                if limit_key is None:
                    UserWarnings.send_warning(Error(exc=RookRuleRateLimited()))
                    return

                msg_id = uuid.uuid4().hex

                logger.info("Executing aug-\t%s (msg ID %s)", self.aug_id, msg_id)
                self._action.execute(self.aug_id, msg_id, namespace, output)
        finally:
            duration = max(utils.get_most_accurate_time_stamp_nano_seconds() - now_ns, RateLimiter.MIN_RATE_LIMIT_VALUE_NS)

            if 0 < self._max_aug_time < duration:
                UserWarnings.set_error(Error(exc=RookRuleMaxExecutionTimeReached()))
                self._enabled = False

            self._rate_limiter.record(limit_key, duration)



import os
import re

GIT_FOLDER = '.git'
GIT_HEAD = 'HEAD'
GIT_CONFIG = 'config'


def is_git(path):
    return os.path.isdir(os.path.join(path, GIT_FOLDER))


def find_root(path):
    if is_git(path):
        return path
    else:
        next_path = os.path.dirname(path)
        if next_path != path:
            return find_root(next_path)
        else:
            return None


def get_revision(path):
    return follow_sym_links(os.path.join(path, GIT_FOLDER), GIT_HEAD)


def get_remote_origin(path):
    with open(os.path.join(path, GIT_FOLDER, GIT_CONFIG)) as config_file:
        config = config_file.read()
    match = re.search(r'\[remote "origin"\]\s*url\s*=\s(?P<url>\S*)', config)
    if match and len(match.groups()) == 1:
        return match.group(0)
    else:
        return ''


def follow_sym_links(root, link):
    with open(os.path.join(root, link), 'r') as f:
        content = f.read()

    if content.startswith("ref:"):
        next_link = content.split(' ')[1].strip()
        return follow_sym_links(root, next_link)
    else:
        return content.strip()



# This file was generated from canopy\ArithmeticPath.peg
# See http://canopy.jcoglan.com/ for documentation.

from collections import defaultdict
import re


class TreeNode(object):
    def __init__(self, text, offset, elements=None):
        self.text = text
        self.offset = offset
        self.elements = elements or []

    def __iter__(self):
        for el in self.elements:
            yield el


class TreeNode1(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode1, self).__init__(text, offset, elements)
        self.atom = elements[0]


class TreeNode2(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode2, self).__init__(text, offset, elements)
        self.opt_ = elements[0]
        self.atom = elements[1]


class TreeNode3(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode3, self).__init__(text, offset, elements)
        self.unicode_set = elements[0]


class TreeNode4(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode4, self).__init__(text, offset, elements)
        self.variable = elements[1]


class TreeNode5(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode5, self).__init__(text, offset, elements)
        self.unicode_set = elements[0]


class TreeNode6(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode6, self).__init__(text, offset, elements)
        self.unicode_set = elements[1]


class TreeNode7(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode7, self).__init__(text, offset, elements)
        self.__ = elements[3]


class TreeNode8(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode8, self).__init__(text, offset, elements)
        self.__ = elements[4]
        self.comp_expression = elements[2]


class TreeNode9(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode9, self).__init__(text, offset, elements)
        self.__ = elements[4]


class TreeNode10(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode10, self).__init__(text, offset, elements)
        self.__ = elements[4]


class TreeNode11(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode11, self).__init__(text, offset, elements)
        self.__ = elements[6]


class TreeNode12(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode12, self).__init__(text, offset, elements)
        self.atom = elements[0]
        self.__ = elements[1]


class TreeNode13(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode13, self).__init__(text, offset, elements)
        self.__ = elements[3]
        self.atom = elements[2]


class TreeNode14(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode14, self).__init__(text, offset, elements)
        self.__ = elements[4]


class TreeNode15(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode15, self).__init__(text, offset, elements)
        self.__ = elements[3]


class TreeNode16(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode16, self).__init__(text, offset, elements)
        self.__ = elements[5]


class TreeNode17(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode17, self).__init__(text, offset, elements)
        self.__ = elements[2]
        self.data = elements[1]


class TreeNode18(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode18, self).__init__(text, offset, elements)
        self.__ = elements[2]
        self.data = elements[1]


class TreeNode19(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode19, self).__init__(text, offset, elements)
        self.__ = elements[2]
        self.data = elements[1]


class TreeNode20(TreeNode):
    def __init__(self, text, offset, elements):
        super(TreeNode20, self).__init__(text, offset, elements)
        self.__ = elements[2]
        self.data = elements[1]


class ParseError(SyntaxError):
    pass


FAILURE = object()


class Grammar(object):
    REGEX_1 = re.compile('^[a-zA-Z_$]')
    REGEX_2 = re.compile('^[a-zA-Z0-9_$]')
    REGEX_3 = re.compile('^[\\"]')
    REGEX_4 = re.compile('^[ !#-~]')
    REGEX_5 = re.compile('^[\\"]')
    REGEX_6 = re.compile('^[^\']')
    REGEX_7 = re.compile('^[\']')
    REGEX_8 = re.compile('^[ !#-~]')
    REGEX_9 = re.compile('^[\']')
    REGEX_10 = re.compile('^[0-9]')
    REGEX_11 = re.compile('^[0-9]')
    REGEX_12 = re.compile('^[0-9]')
    REGEX_13 = re.compile('^[\\s]')

    def _read_comp_expression(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['comp_expression'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read_atom()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            remaining0, index2, elements1, address3 = 0, self._offset, [], True
            while address3 is not FAILURE:
                index3, elements2 = self._offset, []
                address4 = FAILURE
                address4 = self._read_opt_()
                if address4 is not FAILURE:
                    elements2.append(address4)
                    address5 = FAILURE
                    address5 = self._read_atom()
                    if address5 is not FAILURE:
                        elements2.append(address5)
                    else:
                        elements2 = None
                        self._offset = index3
                else:
                    elements2 = None
                    self._offset = index3
                if elements2 is None:
                    address3 = FAILURE
                else:
                    address3 = TreeNode2(self._input[index3:self._offset], index3, elements2)
                    self._offset = self._offset
                if address3 is not FAILURE:
                    elements1.append(address3)
                    remaining0 -= 1
            if remaining0 <= 0:
                address2 = TreeNode(self._input[index2:self._offset], index2, elements1)
                self._offset = self._offset
            else:
                address2 = FAILURE
            if address2 is not FAILURE:
                elements0.append(address2)
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_comp_exp(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['comp_expression'][index0] = (address0, self._offset)
        return address0

    def _read_atom(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['atom'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1 = self._offset
        address0 = self._read_float_()
        if address0 is FAILURE:
            self._offset = index1
            address0 = self._read_number()
            if address0 is FAILURE:
                self._offset = index1
                address0 = self._read_string()
                if address0 is FAILURE:
                    self._offset = index1
                    address0 = self._read_apostrophe_string()
                    if address0 is FAILURE:
                        self._offset = index1
                        address0 = self._read_list()
                        if address0 is FAILURE:
                            self._offset = index1
                            address0 = self._read_chr()
                            if address0 is FAILURE:
                                self._offset = index1
                                address0 = self._read_boolean_()
                                if address0 is FAILURE:
                                    self._offset = index1
                                    address0 = self._read_null_()
                                    if address0 is FAILURE:
                                        self._offset = index1
                                        address0 = self._read_undefined_()
                                        if address0 is FAILURE:
                                            self._offset = index1
                                            address0 = self._read_comp_expression_ex()
                                            if address0 is FAILURE:
                                                self._offset = index1
                                                address0 = self._read_namespace()
                                                if address0 is FAILURE:
                                                    self._offset = index1
        self._cache['atom'][index0] = (address0, self._offset)
        return address0

    def _read_unicode_set(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['unicode_set'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 is not None and Grammar.REGEX_1.search(chunk0):
            address0 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address0 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('[a-zA-Z_$]')
        self._cache['unicode_set'][index0] = (address0, self._offset)
        return address0

    def _read_unicode_set_with_numbers(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['unicode_set_with_numbers'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 is not None and Grammar.REGEX_2.search(chunk0):
            address0 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address0 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('[a-zA-Z0-9_$]')
        self._cache['unicode_set_with_numbers'][index0] = (address0, self._offset)
        return address0

    def _read_variable(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['variable'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read_unicode_set()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            remaining0, index2, elements1, address3 = 0, self._offset, [], True
            while address3 is not FAILURE:
                address3 = self._read_unicode_set_with_numbers()
                if address3 is not FAILURE:
                    elements1.append(address3)
                    remaining0 -= 1
            if remaining0 <= 0:
                address2 = TreeNode(self._input[index2:self._offset], index2, elements1)
                self._offset = self._offset
            else:
                address2 = FAILURE
            if address2 is not FAILURE:
                elements0.append(address2)
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_attribute(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['variable'][index0] = (address0, self._offset)
        return address0

    def _read_variable_access(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['variable_access'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 == '.':
            address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address1 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('"."')
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            address2 = self._read_variable()
            if address2 is not FAILURE:
                elements0.append(address2)
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_attribute_operation(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['variable_access'][index0] = (address0, self._offset)
        return address0

    def _read_map_access(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['map_access'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 == '[':
            address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address1 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('"["')
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            address2 = self._read_chr()
            if address2 is FAILURE:
                self._offset = index2
                address2 = self._read_string()
                if address2 is FAILURE:
                    self._offset = index2
                    address2 = self._read_apostrophe_string()
                    if address2 is FAILURE:
                        self._offset = index2
                        address2 = self._read_number()
                        if address2 is FAILURE:
                            self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 1]
                if chunk1 == ']':
                    address3 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address3 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"]"')
                if address3 is not FAILURE:
                    elements0.append(address3)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_lookup_operation(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['map_access'][index0] = (address0, self._offset)
        return address0

    def _read_func_call(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['func_call'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read_unicode_set()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            remaining0, index2, elements1, address3 = 0, self._offset, [], True
            while address3 is not FAILURE:
                address3 = self._read_unicode_set_with_numbers()
                if address3 is not FAILURE:
                    elements1.append(address3)
                    remaining0 -= 1
            if remaining0 <= 0:
                address2 = TreeNode(self._input[index2:self._offset], index2, elements1)
                self._offset = self._offset
            else:
                address2 = FAILURE
            if address2 is not FAILURE:
                elements0.append(address2)
                address4 = FAILURE
                chunk0 = None
                if self._offset < self._input_size:
                    chunk0 = self._input[self._offset:self._offset + 1]
                if chunk0 == '(':
                    address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address4 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"("')
                if address4 is not FAILURE:
                    elements0.append(address4)
                    address5 = FAILURE
                    index3 = self._offset
                    address5 = self._read_atom()
                    if address5 is FAILURE:
                        address5 = TreeNode(self._input[index3:index3], index3)
                        self._offset = index3
                    if address5 is not FAILURE:
                        elements0.append(address5)
                        address6 = FAILURE
                        chunk1 = None
                        if self._offset < self._input_size:
                            chunk1 = self._input[self._offset:self._offset + 1]
                        if chunk1 == ')':
                            address6 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                            self._offset = self._offset + 1
                        else:
                            address6 = FAILURE
                            if self._offset > self._failure:
                                self._failure = self._offset
                                self._expected = []
                            if self._offset == self._failure:
                                self._expected.append('")"')
                        if address6 is not FAILURE:
                            elements0.append(address6)
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_function_operation(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['func_call'][index0] = (address0, self._offset)
        return address0

    def _read_func_call_access(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['func_call_access'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        chunk0 = None
        if self._offset < self._input_size:
            chunk0 = self._input[self._offset:self._offset + 1]
        if chunk0 == '.':
            address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
            self._offset = self._offset + 1
        else:
            address1 = FAILURE
            if self._offset > self._failure:
                self._failure = self._offset
                self._expected = []
            if self._offset == self._failure:
                self._expected.append('"."')
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            address2 = self._read_unicode_set()
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                remaining0, index2, elements1, address4 = 0, self._offset, [], True
                while address4 is not FAILURE:
                    address4 = self._read_unicode_set_with_numbers()
                    if address4 is not FAILURE:
                        elements1.append(address4)
                        remaining0 -= 1
                if remaining0 <= 0:
                    address3 = TreeNode(self._input[index2:self._offset], index2, elements1)
                    self._offset = self._offset
                else:
                    address3 = FAILURE
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address5 = FAILURE
                    chunk1 = None
                    if self._offset < self._input_size:
                        chunk1 = self._input[self._offset:self._offset + 1]
                    if chunk1 == '(':
                        address5 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address5 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('"("')
                    if address5 is not FAILURE:
                        elements0.append(address5)
                        address6 = FAILURE
                        index3 = self._offset
                        address6 = self._read_atom()
                        if address6 is FAILURE:
                            address6 = TreeNode(self._input[index3:index3], index3)
                            self._offset = index3
                        if address6 is not FAILURE:
                            elements0.append(address6)
                            address7 = FAILURE
                            chunk2 = None
                            if self._offset < self._input_size:
                                chunk2 = self._input[self._offset:self._offset + 1]
                            if chunk2 == ')':
                                address7 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                                self._offset = self._offset + 1
                            else:
                                address7 = FAILURE
                                if self._offset > self._failure:
                                    self._failure = self._offset
                                    self._expected = []
                                if self._offset == self._failure:
                                    self._expected.append('")"')
                            if address7 is not FAILURE:
                                elements0.append(address7)
                            else:
                                elements0 = None
                                self._offset = index1
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_function_operation_access(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['func_call_access'][index0] = (address0, self._offset)
        return address0

    def _read_namespace(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['namespace'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            address2 = self._read_func_call()
            if address2 is FAILURE:
                self._offset = index2
                address2 = self._read_variable()
                if address2 is FAILURE:
                    self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                remaining0, index3, elements1, address4 = 0, self._offset, [], True
                while address4 is not FAILURE:
                    index4 = self._offset
                    address4 = self._read_func_call_access()
                    if address4 is FAILURE:
                        self._offset = index4
                        address4 = self._read_map_access()
                        if address4 is FAILURE:
                            self._offset = index4
                            address4 = self._read_variable_access()
                            if address4 is FAILURE:
                                self._offset = index4
                    if address4 is not FAILURE:
                        elements1.append(address4)
                        remaining0 -= 1
                if remaining0 <= 0:
                    address3 = TreeNode(self._input[index3:self._offset], index3, elements1)
                    self._offset = self._offset
                else:
                    address3 = FAILURE
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address5 = FAILURE
                    address5 = self._read___()
                    if address5 is not FAILURE:
                        elements0.append(address5)
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_and_execute_namespace_operation(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['namespace'][index0] = (address0, self._offset)
        return address0

    def _read_comp_expression_ex(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['comp_expression_ex'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 == '(':
                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"("')
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                address3 = self._read_comp_expression()
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address4 = FAILURE
                    chunk1 = None
                    if self._offset < self._input_size:
                        chunk1 = self._input[self._offset:self._offset + 1]
                    if chunk1 == ')':
                        address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address4 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('")"')
                    if address4 is not FAILURE:
                        elements0.append(address4)
                        address5 = FAILURE
                        address5 = self._read___()
                        if address5 is not FAILURE:
                            elements0.append(address5)
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_comp_exp_ex(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['comp_expression_ex'][index0] = (address0, self._offset)
        return address0

    def _read_string(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['string'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 is not None and Grammar.REGEX_3.search(chunk0):
                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('[\\"]')
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                remaining0, index2, elements1, address4 = 0, self._offset, [], True
                while address4 is not FAILURE:
                    chunk1 = None
                    if self._offset < self._input_size:
                        chunk1 = self._input[self._offset:self._offset + 1]
                    if chunk1 is not None and Grammar.REGEX_4.search(chunk1):
                        address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address4 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('[ !#-~]')
                    if address4 is not FAILURE:
                        elements1.append(address4)
                        remaining0 -= 1
                if remaining0 <= 0:
                    address3 = TreeNode(self._input[index2:self._offset], index2, elements1)
                    self._offset = self._offset
                else:
                    address3 = FAILURE
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address5 = FAILURE
                    chunk2 = None
                    if self._offset < self._input_size:
                        chunk2 = self._input[self._offset:self._offset + 1]
                    if chunk2 is not None and Grammar.REGEX_5.search(chunk2):
                        address5 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address5 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('[\\"]')
                    if address5 is not FAILURE:
                        elements0.append(address5)
                        address6 = FAILURE
                        address6 = self._read___()
                        if address6 is not FAILURE:
                            elements0.append(address6)
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_string(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['string'][index0] = (address0, self._offset)
        return address0

    def _read_apostrophe_string(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['apostrophe_string'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 == '\'':
                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"\'"')
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                remaining0, index2, elements1, address4 = 0, self._offset, [], True
                while address4 is not FAILURE:
                    chunk1 = None
                    if self._offset < self._input_size:
                        chunk1 = self._input[self._offset:self._offset + 1]
                    if chunk1 is not None and Grammar.REGEX_6.search(chunk1):
                        address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address4 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('[^\']')
                    if address4 is not FAILURE:
                        elements1.append(address4)
                        remaining0 -= 1
                if remaining0 <= 0:
                    address3 = TreeNode(self._input[index2:self._offset], index2, elements1)
                    self._offset = self._offset
                else:
                    address3 = FAILURE
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address5 = FAILURE
                    chunk2 = None
                    if self._offset < self._input_size:
                        chunk2 = self._input[self._offset:self._offset + 1]
                    if chunk2 == '\'':
                        address5 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address5 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('"\'"')
                    if address5 is not FAILURE:
                        elements0.append(address5)
                        address6 = FAILURE
                        address6 = self._read___()
                        if address6 is not FAILURE:
                            elements0.append(address6)
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_apostrophe_string(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['apostrophe_string'][index0] = (address0, self._offset)
        return address0

    def _read_list(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['list'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 == '[':
                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"["')
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                address3 = self._read___()
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address4 = FAILURE
                    index2 = self._offset
                    index3, elements1 = self._offset, []
                    address5 = FAILURE
                    address5 = self._read_atom()
                    if address5 is not FAILURE:
                        elements1.append(address5)
                        address6 = FAILURE
                        address6 = self._read___()
                        if address6 is not FAILURE:
                            elements1.append(address6)
                            address7 = FAILURE
                            remaining0, index4, elements2, address8 = 0, self._offset, [], True
                            while address8 is not FAILURE:
                                index5, elements3 = self._offset, []
                                address9 = FAILURE
                                chunk1 = None
                                if self._offset < self._input_size:
                                    chunk1 = self._input[self._offset:self._offset + 1]
                                if chunk1 == ',':
                                    address9 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                                    self._offset = self._offset + 1
                                else:
                                    address9 = FAILURE
                                    if self._offset > self._failure:
                                        self._failure = self._offset
                                        self._expected = []
                                    if self._offset == self._failure:
                                        self._expected.append('","')
                                if address9 is not FAILURE:
                                    elements3.append(address9)
                                    address10 = FAILURE
                                    address10 = self._read___()
                                    if address10 is not FAILURE:
                                        elements3.append(address10)
                                        address11 = FAILURE
                                        address11 = self._read_atom()
                                        if address11 is not FAILURE:
                                            elements3.append(address11)
                                            address12 = FAILURE
                                            address12 = self._read___()
                                            if address12 is not FAILURE:
                                                elements3.append(address12)
                                            else:
                                                elements3 = None
                                                self._offset = index5
                                        else:
                                            elements3 = None
                                            self._offset = index5
                                    else:
                                        elements3 = None
                                        self._offset = index5
                                else:
                                    elements3 = None
                                    self._offset = index5
                                if elements3 is None:
                                    address8 = FAILURE
                                else:
                                    address8 = TreeNode13(self._input[index5:self._offset], index5, elements3)
                                    self._offset = self._offset
                                if address8 is not FAILURE:
                                    elements2.append(address8)
                                    remaining0 -= 1
                            if remaining0 <= 0:
                                address7 = TreeNode(self._input[index4:self._offset], index4, elements2)
                                self._offset = self._offset
                            else:
                                address7 = FAILURE
                            if address7 is not FAILURE:
                                elements1.append(address7)
                            else:
                                elements1 = None
                                self._offset = index3
                        else:
                            elements1 = None
                            self._offset = index3
                    else:
                        elements1 = None
                        self._offset = index3
                    if elements1 is None:
                        address4 = FAILURE
                    else:
                        address4 = TreeNode12(self._input[index3:self._offset], index3, elements1)
                        self._offset = self._offset
                    if address4 is FAILURE:
                        address4 = TreeNode(self._input[index2:index2], index2)
                        self._offset = index2
                    if address4 is not FAILURE:
                        elements0.append(address4)
                        address13 = FAILURE
                        address13 = self._read___()
                        if address13 is not FAILURE:
                            elements0.append(address13)
                            address14 = FAILURE
                            chunk2 = None
                            if self._offset < self._input_size:
                                chunk2 = self._input[self._offset:self._offset + 1]
                            if chunk2 == ']':
                                address14 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                                self._offset = self._offset + 1
                            else:
                                address14 = FAILURE
                                if self._offset > self._failure:
                                    self._failure = self._offset
                                    self._expected = []
                                if self._offset == self._failure:
                                    self._expected.append('"]"')
                            if address14 is not FAILURE:
                                elements0.append(address14)
                                address15 = FAILURE
                                address15 = self._read___()
                                if address15 is not FAILURE:
                                    elements0.append(address15)
                                else:
                                    elements0 = None
                                    self._offset = index1
                            else:
                                elements0 = None
                                self._offset = index1
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_list(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['list'][index0] = (address0, self._offset)
        return address0

    def _read_chr(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['chr'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 is not None and Grammar.REGEX_7.search(chunk0):
                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('[\']')
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 1]
                if chunk1 is not None and Grammar.REGEX_8.search(chunk1):
                    address3 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address3 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('[ !#-~]')
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address4 = FAILURE
                    chunk2 = None
                    if self._offset < self._input_size:
                        chunk2 = self._input[self._offset:self._offset + 1]
                    if chunk2 is not None and Grammar.REGEX_9.search(chunk2):
                        address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address4 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('[\']')
                    if address4 is not FAILURE:
                        elements0.append(address4)
                        address5 = FAILURE
                        address5 = self._read___()
                        if address5 is not FAILURE:
                            elements0.append(address5)
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_char(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['chr'][index0] = (address0, self._offset)
        return address0

    def _read_number(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['number'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 == '-':
                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"-"')
            if address2 is FAILURE:
                address2 = TreeNode(self._input[index2:index2], index2)
                self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                remaining0, index3, elements1, address4 = 1, self._offset, [], True
                while address4 is not FAILURE:
                    chunk1 = None
                    if self._offset < self._input_size:
                        chunk1 = self._input[self._offset:self._offset + 1]
                    if chunk1 is not None and Grammar.REGEX_10.search(chunk1):
                        address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address4 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('[0-9]')
                    if address4 is not FAILURE:
                        elements1.append(address4)
                        remaining0 -= 1
                if remaining0 <= 0:
                    address3 = TreeNode(self._input[index3:self._offset], index3, elements1)
                    self._offset = self._offset
                else:
                    address3 = FAILURE
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address5 = FAILURE
                    address5 = self._read___()
                    if address5 is not FAILURE:
                        elements0.append(address5)
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_number(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['number'][index0] = (address0, self._offset)
        return address0

    def _read_float_(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['float_'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 == '-':
                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"-"')
            if address2 is FAILURE:
                address2 = TreeNode(self._input[index2:index2], index2)
                self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                remaining0, index3, elements1, address4 = 1, self._offset, [], True
                while address4 is not FAILURE:
                    chunk1 = None
                    if self._offset < self._input_size:
                        chunk1 = self._input[self._offset:self._offset + 1]
                    if chunk1 is not None and Grammar.REGEX_11.search(chunk1):
                        address4 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address4 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('[0-9]')
                    if address4 is not FAILURE:
                        elements1.append(address4)
                        remaining0 -= 1
                if remaining0 <= 0:
                    address3 = TreeNode(self._input[index3:self._offset], index3, elements1)
                    self._offset = self._offset
                else:
                    address3 = FAILURE
                if address3 is not FAILURE:
                    elements0.append(address3)
                    address5 = FAILURE
                    chunk2 = None
                    if self._offset < self._input_size:
                        chunk2 = self._input[self._offset:self._offset + 1]
                    if chunk2 == '.':
                        address5 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address5 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('"."')
                    if address5 is not FAILURE:
                        elements0.append(address5)
                        address6 = FAILURE
                        remaining1, index4, elements2, address7 = 1, self._offset, [], True
                        while address7 is not FAILURE:
                            chunk3 = None
                            if self._offset < self._input_size:
                                chunk3 = self._input[self._offset:self._offset + 1]
                            if chunk3 is not None and Grammar.REGEX_12.search(chunk3):
                                address7 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                                self._offset = self._offset + 1
                            else:
                                address7 = FAILURE
                                if self._offset > self._failure:
                                    self._failure = self._offset
                                    self._expected = []
                                if self._offset == self._failure:
                                    self._expected.append('[0-9]')
                            if address7 is not FAILURE:
                                elements2.append(address7)
                                remaining1 -= 1
                        if remaining1 <= 0:
                            address6 = TreeNode(self._input[index4:self._offset], index4, elements2)
                            self._offset = self._offset
                        else:
                            address6 = FAILURE
                        if address6 is not FAILURE:
                            elements0.append(address6)
                            address8 = FAILURE
                            address8 = self._read___()
                            if address8 is not FAILURE:
                                elements0.append(address8)
                            else:
                                elements0 = None
                                self._offset = index1
                        else:
                            elements0 = None
                            self._offset = index1
                    else:
                        elements0 = None
                        self._offset = index1
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_float(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['float_'][index0] = (address0, self._offset)
        return address0

    def _read_boolean_(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['boolean_'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 4]
            if chunk0 == 'True':
                address2 = TreeNode(self._input[self._offset:self._offset + 4], self._offset)
                self._offset = self._offset + 4
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"True"')
            if address2 is FAILURE:
                self._offset = index2
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 4]
                if chunk1 == 'true':
                    address2 = TreeNode(self._input[self._offset:self._offset + 4], self._offset)
                    self._offset = self._offset + 4
                else:
                    address2 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"true"')
                if address2 is FAILURE:
                    self._offset = index2
                    chunk2 = None
                    if self._offset < self._input_size:
                        chunk2 = self._input[self._offset:self._offset + 5]
                    if chunk2 == 'False':
                        address2 = TreeNode(self._input[self._offset:self._offset + 5], self._offset)
                        self._offset = self._offset + 5
                    else:
                        address2 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('"False"')
                    if address2 is FAILURE:
                        self._offset = index2
                        chunk3 = None
                        if self._offset < self._input_size:
                            chunk3 = self._input[self._offset:self._offset + 5]
                        if chunk3 == 'false':
                            address2 = TreeNode(self._input[self._offset:self._offset + 5], self._offset)
                            self._offset = self._offset + 5
                        else:
                            address2 = FAILURE
                            if self._offset > self._failure:
                                self._failure = self._offset
                                self._expected = []
                            if self._offset == self._failure:
                                self._expected.append('"false"')
                        if address2 is FAILURE:
                            self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                address3 = self._read___()
                if address3 is not FAILURE:
                    elements0.append(address3)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_bool(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['boolean_'][index0] = (address0, self._offset)
        return address0

    def _read_null_(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['null_'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 4]
            if chunk0 == 'None':
                address2 = TreeNode(self._input[self._offset:self._offset + 4], self._offset)
                self._offset = self._offset + 4
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"None"')
            if address2 is FAILURE:
                self._offset = index2
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 3]
                if chunk1 == 'nil':
                    address2 = TreeNode(self._input[self._offset:self._offset + 3], self._offset)
                    self._offset = self._offset + 3
                else:
                    address2 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"nil"')
                if address2 is FAILURE:
                    self._offset = index2
                    chunk2 = None
                    if self._offset < self._input_size:
                        chunk2 = self._input[self._offset:self._offset + 4]
                    if chunk2 == 'null':
                        address2 = TreeNode(self._input[self._offset:self._offset + 4], self._offset)
                        self._offset = self._offset + 4
                    else:
                        address2 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('"null"')
                    if address2 is FAILURE:
                        self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                address3 = self._read___()
                if address3 is not FAILURE:
                    elements0.append(address3)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_null(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['null_'][index0] = (address0, self._offset)
        return address0

    def _read_undefined_(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['undefined_'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 9]
            if chunk0 == 'undefined':
                address2 = TreeNode(self._input[self._offset:self._offset + 9], self._offset)
                self._offset = self._offset + 9
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"undefined"')
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                address3 = self._read___()
                if address3 is not FAILURE:
                    elements0.append(address3)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_undefined(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['undefined_'][index0] = (address0, self._offset)
        return address0

    def _read_opt_(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['opt_'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        index1, elements0 = self._offset, []
        address1 = FAILURE
        address1 = self._read___()
        if address1 is not FAILURE:
            elements0.append(address1)
            address2 = FAILURE
            index2 = self._offset
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 == '+':
                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address2 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('"+"')
            if address2 is FAILURE:
                self._offset = index2
                chunk1 = None
                if self._offset < self._input_size:
                    chunk1 = self._input[self._offset:self._offset + 1]
                if chunk1 == '-':
                    address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                    self._offset = self._offset + 1
                else:
                    address2 = FAILURE
                    if self._offset > self._failure:
                        self._failure = self._offset
                        self._expected = []
                    if self._offset == self._failure:
                        self._expected.append('"-"')
                if address2 is FAILURE:
                    self._offset = index2
                    chunk2 = None
                    if self._offset < self._input_size:
                        chunk2 = self._input[self._offset:self._offset + 1]
                    if chunk2 == '*':
                        address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                        self._offset = self._offset + 1
                    else:
                        address2 = FAILURE
                        if self._offset > self._failure:
                            self._failure = self._offset
                            self._expected = []
                        if self._offset == self._failure:
                            self._expected.append('"*"')
                    if address2 is FAILURE:
                        self._offset = index2
                        chunk3 = None
                        if self._offset < self._input_size:
                            chunk3 = self._input[self._offset:self._offset + 1]
                        if chunk3 == '/':
                            address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                            self._offset = self._offset + 1
                        else:
                            address2 = FAILURE
                            if self._offset > self._failure:
                                self._failure = self._offset
                                self._expected = []
                            if self._offset == self._failure:
                                self._expected.append('"/"')
                        if address2 is FAILURE:
                            self._offset = index2
                            chunk4 = None
                            if self._offset < self._input_size:
                                chunk4 = self._input[self._offset:self._offset + 2]
                            if chunk4 == '<=':
                                address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                self._offset = self._offset + 2
                            else:
                                address2 = FAILURE
                                if self._offset > self._failure:
                                    self._failure = self._offset
                                    self._expected = []
                                if self._offset == self._failure:
                                    self._expected.append('"<="')
                            if address2 is FAILURE:
                                self._offset = index2
                                chunk5 = None
                                if self._offset < self._input_size:
                                    chunk5 = self._input[self._offset:self._offset + 2]
                                if chunk5 == '>=':
                                    address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                    self._offset = self._offset + 2
                                else:
                                    address2 = FAILURE
                                    if self._offset > self._failure:
                                        self._failure = self._offset
                                        self._expected = []
                                    if self._offset == self._failure:
                                        self._expected.append('">="')
                                if address2 is FAILURE:
                                    self._offset = index2
                                    chunk6 = None
                                    if self._offset < self._input_size:
                                        chunk6 = self._input[self._offset:self._offset + 2]
                                    if chunk6 == '!=':
                                        address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                        self._offset = self._offset + 2
                                    else:
                                        address2 = FAILURE
                                        if self._offset > self._failure:
                                            self._failure = self._offset
                                            self._expected = []
                                        if self._offset == self._failure:
                                            self._expected.append('"!="')
                                    if address2 is FAILURE:
                                        self._offset = index2
                                        chunk7 = None
                                        if self._offset < self._input_size:
                                            chunk7 = self._input[self._offset:self._offset + 2]
                                        if chunk7 == '==':
                                            address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                            self._offset = self._offset + 2
                                        else:
                                            address2 = FAILURE
                                            if self._offset > self._failure:
                                                self._failure = self._offset
                                                self._expected = []
                                            if self._offset == self._failure:
                                                self._expected.append('"=="')
                                        if address2 is FAILURE:
                                            self._offset = index2
                                            chunk8 = None
                                            if self._offset < self._input_size:
                                                chunk8 = self._input[self._offset:self._offset + 1]
                                            if chunk8 == '=':
                                                address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                                                self._offset = self._offset + 1
                                            else:
                                                address2 = FAILURE
                                                if self._offset > self._failure:
                                                    self._failure = self._offset
                                                    self._expected = []
                                                if self._offset == self._failure:
                                                    self._expected.append('"="')
                                            if address2 is FAILURE:
                                                self._offset = index2
                                                chunk9 = None
                                                if self._offset < self._input_size:
                                                    chunk9 = self._input[self._offset:self._offset + 1]
                                                if chunk9 == '>':
                                                    address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                                                    self._offset = self._offset + 1
                                                else:
                                                    address2 = FAILURE
                                                    if self._offset > self._failure:
                                                        self._failure = self._offset
                                                        self._expected = []
                                                    if self._offset == self._failure:
                                                        self._expected.append('">"')
                                                if address2 is FAILURE:
                                                    self._offset = index2
                                                    chunk10 = None
                                                    if self._offset < self._input_size:
                                                        chunk10 = self._input[self._offset:self._offset + 1]
                                                    if chunk10 == '<':
                                                        address2 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                                                        self._offset = self._offset + 1
                                                    else:
                                                        address2 = FAILURE
                                                        if self._offset > self._failure:
                                                            self._failure = self._offset
                                                            self._expected = []
                                                        if self._offset == self._failure:
                                                            self._expected.append('"<"')
                                                    if address2 is FAILURE:
                                                        self._offset = index2
                                                        chunk11 = None
                                                        if self._offset < self._input_size:
                                                            chunk11 = self._input[self._offset:self._offset + 2]
                                                        if chunk11 == 'LT':
                                                            address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                            self._offset = self._offset + 2
                                                        else:
                                                            address2 = FAILURE
                                                            if self._offset > self._failure:
                                                                self._failure = self._offset
                                                                self._expected = []
                                                            if self._offset == self._failure:
                                                                self._expected.append('"LT"')
                                                        if address2 is FAILURE:
                                                            self._offset = index2
                                                            chunk12 = None
                                                            if self._offset < self._input_size:
                                                                chunk12 = self._input[self._offset:self._offset + 2]
                                                            if chunk12 == 'GT':
                                                                address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                self._offset = self._offset + 2
                                                            else:
                                                                address2 = FAILURE
                                                                if self._offset > self._failure:
                                                                    self._failure = self._offset
                                                                    self._expected = []
                                                                if self._offset == self._failure:
                                                                    self._expected.append('"GT"')
                                                            if address2 is FAILURE:
                                                                self._offset = index2
                                                                chunk13 = None
                                                                if self._offset < self._input_size:
                                                                    chunk13 = self._input[self._offset:self._offset + 2]
                                                                if chunk13 == 'LE':
                                                                    address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                    self._offset = self._offset + 2
                                                                else:
                                                                    address2 = FAILURE
                                                                    if self._offset > self._failure:
                                                                        self._failure = self._offset
                                                                        self._expected = []
                                                                    if self._offset == self._failure:
                                                                        self._expected.append('"LE"')
                                                                if address2 is FAILURE:
                                                                    self._offset = index2
                                                                    chunk14 = None
                                                                    if self._offset < self._input_size:
                                                                        chunk14 = self._input[self._offset:self._offset + 2]
                                                                    if chunk14 == 'GE':
                                                                        address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                        self._offset = self._offset + 2
                                                                    else:
                                                                        address2 = FAILURE
                                                                        if self._offset > self._failure:
                                                                            self._failure = self._offset
                                                                            self._expected = []
                                                                        if self._offset == self._failure:
                                                                            self._expected.append('"GE"')
                                                                    if address2 is FAILURE:
                                                                        self._offset = index2
                                                                        chunk15 = None
                                                                        if self._offset < self._input_size:
                                                                            chunk15 = self._input[self._offset:self._offset + 2]
                                                                        if chunk15 == 'EQ':
                                                                            address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                            self._offset = self._offset + 2
                                                                        else:
                                                                            address2 = FAILURE
                                                                            if self._offset > self._failure:
                                                                                self._failure = self._offset
                                                                                self._expected = []
                                                                            if self._offset == self._failure:
                                                                                self._expected.append('"EQ"')
                                                                        if address2 is FAILURE:
                                                                            self._offset = index2
                                                                            chunk16 = None
                                                                            if self._offset < self._input_size:
                                                                                chunk16 = self._input[self._offset:self._offset + 2]
                                                                            if chunk16 == 'NE':
                                                                                address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                self._offset = self._offset + 2
                                                                            else:
                                                                                address2 = FAILURE
                                                                                if self._offset > self._failure:
                                                                                    self._failure = self._offset
                                                                                    self._expected = []
                                                                                if self._offset == self._failure:
                                                                                    self._expected.append('"NE"')
                                                                            if address2 is FAILURE:
                                                                                self._offset = index2
                                                                                chunk17 = None
                                                                                if self._offset < self._input_size:
                                                                                    chunk17 = self._input[self._offset:self._offset + 2]
                                                                                if chunk17 == 'lt':
                                                                                    address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                    self._offset = self._offset + 2
                                                                                else:
                                                                                    address2 = FAILURE
                                                                                    if self._offset > self._failure:
                                                                                        self._failure = self._offset
                                                                                        self._expected = []
                                                                                    if self._offset == self._failure:
                                                                                        self._expected.append('"lt"')
                                                                                if address2 is FAILURE:
                                                                                    self._offset = index2
                                                                                    chunk18 = None
                                                                                    if self._offset < self._input_size:
                                                                                        chunk18 = self._input[self._offset:self._offset + 2]
                                                                                    if chunk18 == 'gt':
                                                                                        address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                        self._offset = self._offset + 2
                                                                                    else:
                                                                                        address2 = FAILURE
                                                                                        if self._offset > self._failure:
                                                                                            self._failure = self._offset
                                                                                            self._expected = []
                                                                                        if self._offset == self._failure:
                                                                                            self._expected.append('"gt"')
                                                                                    if address2 is FAILURE:
                                                                                        self._offset = index2
                                                                                        chunk19 = None
                                                                                        if self._offset < self._input_size:
                                                                                            chunk19 = self._input[self._offset:self._offset + 2]
                                                                                        if chunk19 == 'le':
                                                                                            address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                            self._offset = self._offset + 2
                                                                                        else:
                                                                                            address2 = FAILURE
                                                                                            if self._offset > self._failure:
                                                                                                self._failure = self._offset
                                                                                                self._expected = []
                                                                                            if self._offset == self._failure:
                                                                                                self._expected.append('"le"')
                                                                                        if address2 is FAILURE:
                                                                                            self._offset = index2
                                                                                            chunk20 = None
                                                                                            if self._offset < self._input_size:
                                                                                                chunk20 = self._input[self._offset:self._offset + 2]
                                                                                            if chunk20 == 'ge':
                                                                                                address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                self._offset = self._offset + 2
                                                                                            else:
                                                                                                address2 = FAILURE
                                                                                                if self._offset > self._failure:
                                                                                                    self._failure = self._offset
                                                                                                    self._expected = []
                                                                                                if self._offset == self._failure:
                                                                                                    self._expected.append('"ge"')
                                                                                            if address2 is FAILURE:
                                                                                                self._offset = index2
                                                                                                chunk21 = None
                                                                                                if self._offset < self._input_size:
                                                                                                    chunk21 = self._input[self._offset:self._offset + 2]
                                                                                                if chunk21 == 'eq':
                                                                                                    address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                    self._offset = self._offset + 2
                                                                                                else:
                                                                                                    address2 = FAILURE
                                                                                                    if self._offset > self._failure:
                                                                                                        self._failure = self._offset
                                                                                                        self._expected = []
                                                                                                    if self._offset == self._failure:
                                                                                                        self._expected.append('"eq"')
                                                                                                if address2 is FAILURE:
                                                                                                    self._offset = index2
                                                                                                    chunk22 = None
                                                                                                    if self._offset < self._input_size:
                                                                                                        chunk22 = self._input[self._offset:self._offset + 2]
                                                                                                    if chunk22 == 'ne':
                                                                                                        address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                        self._offset = self._offset + 2
                                                                                                    else:
                                                                                                        address2 = FAILURE
                                                                                                        if self._offset > self._failure:
                                                                                                            self._failure = self._offset
                                                                                                            self._expected = []
                                                                                                        if self._offset == self._failure:
                                                                                                            self._expected.append('"ne"')
                                                                                                    if address2 is FAILURE:
                                                                                                        self._offset = index2
                                                                                                        chunk23 = None
                                                                                                        if self._offset < self._input_size:
                                                                                                            chunk23 = self._input[self._offset:self._offset + 2]
                                                                                                        if chunk23 == 'in':
                                                                                                            address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                            self._offset = self._offset + 2
                                                                                                        else:
                                                                                                            address2 = FAILURE
                                                                                                            if self._offset > self._failure:
                                                                                                                self._failure = self._offset
                                                                                                                self._expected = []
                                                                                                            if self._offset == self._failure:
                                                                                                                self._expected.append('"in"')
                                                                                                        if address2 is FAILURE:
                                                                                                            self._offset = index2
                                                                                                            chunk24 = None
                                                                                                            if self._offset < self._input_size:
                                                                                                                chunk24 = self._input[self._offset:self._offset + 2]
                                                                                                            if chunk24 == 'IN':
                                                                                                                address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                                self._offset = self._offset + 2
                                                                                                            else:
                                                                                                                address2 = FAILURE
                                                                                                                if self._offset > self._failure:
                                                                                                                    self._failure = self._offset
                                                                                                                    self._expected = []
                                                                                                                if self._offset == self._failure:
                                                                                                                    self._expected.append('"IN"')
                                                                                                            if address2 is FAILURE:
                                                                                                                self._offset = index2
                                                                                                                chunk25 = None
                                                                                                                if self._offset < self._input_size:
                                                                                                                    chunk25 = self._input[self._offset:self._offset + 2]
                                                                                                                if chunk25 == 'or':
                                                                                                                    address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                                    self._offset = self._offset + 2
                                                                                                                else:
                                                                                                                    address2 = FAILURE
                                                                                                                    if self._offset > self._failure:
                                                                                                                        self._failure = self._offset
                                                                                                                        self._expected = []
                                                                                                                    if self._offset == self._failure:
                                                                                                                        self._expected.append('"or"')
                                                                                                                if address2 is FAILURE:
                                                                                                                    self._offset = index2
                                                                                                                    chunk26 = None
                                                                                                                    if self._offset < self._input_size:
                                                                                                                        chunk26 = self._input[self._offset:self._offset + 2]
                                                                                                                    if chunk26 == 'OR':
                                                                                                                        address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                                        self._offset = self._offset + 2
                                                                                                                    else:
                                                                                                                        address2 = FAILURE
                                                                                                                        if self._offset > self._failure:
                                                                                                                            self._failure = self._offset
                                                                                                                            self._expected = []
                                                                                                                        if self._offset == self._failure:
                                                                                                                            self._expected.append('"OR"')
                                                                                                                    if address2 is FAILURE:
                                                                                                                        self._offset = index2
                                                                                                                        chunk27 = None
                                                                                                                        if self._offset < self._input_size:
                                                                                                                            chunk27 = self._input[self._offset:self._offset + 2]
                                                                                                                        if chunk27 == '||':
                                                                                                                            address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                                            self._offset = self._offset + 2
                                                                                                                        else:
                                                                                                                            address2 = FAILURE
                                                                                                                            if self._offset > self._failure:
                                                                                                                                self._failure = self._offset
                                                                                                                                self._expected = []
                                                                                                                            if self._offset == self._failure:
                                                                                                                                self._expected.append('"||"')
                                                                                                                        if address2 is FAILURE:
                                                                                                                            self._offset = index2
                                                                                                                            chunk28 = None
                                                                                                                            if self._offset < self._input_size:
                                                                                                                                chunk28 = self._input[self._offset:self._offset + 3]
                                                                                                                            if chunk28 == 'and':
                                                                                                                                address2 = TreeNode(self._input[self._offset:self._offset + 3], self._offset)
                                                                                                                                self._offset = self._offset + 3
                                                                                                                            else:
                                                                                                                                address2 = FAILURE
                                                                                                                                if self._offset > self._failure:
                                                                                                                                    self._failure = self._offset
                                                                                                                                    self._expected = []
                                                                                                                                if self._offset == self._failure:
                                                                                                                                    self._expected.append('"and"')
                                                                                                                            if address2 is FAILURE:
                                                                                                                                self._offset = index2
                                                                                                                                chunk29 = None
                                                                                                                                if self._offset < self._input_size:
                                                                                                                                    chunk29 = self._input[self._offset:self._offset + 3]
                                                                                                                                if chunk29 == 'AND':
                                                                                                                                    address2 = TreeNode(self._input[self._offset:self._offset + 3], self._offset)
                                                                                                                                    self._offset = self._offset + 3
                                                                                                                                else:
                                                                                                                                    address2 = FAILURE
                                                                                                                                    if self._offset > self._failure:
                                                                                                                                        self._failure = self._offset
                                                                                                                                        self._expected = []
                                                                                                                                    if self._offset == self._failure:
                                                                                                                                        self._expected.append('"AND"')
                                                                                                                                if address2 is FAILURE:
                                                                                                                                    self._offset = index2
                                                                                                                                    chunk30 = None
                                                                                                                                    if self._offset < self._input_size:
                                                                                                                                        chunk30 = self._input[self._offset:self._offset + 2]
                                                                                                                                    if chunk30 == '&&':
                                                                                                                                        address2 = TreeNode(self._input[self._offset:self._offset + 2], self._offset)
                                                                                                                                        self._offset = self._offset + 2
                                                                                                                                    else:
                                                                                                                                        address2 = FAILURE
                                                                                                                                        if self._offset > self._failure:
                                                                                                                                            self._failure = self._offset
                                                                                                                                            self._expected = []
                                                                                                                                        if self._offset == self._failure:
                                                                                                                                            self._expected.append('"&&"')
                                                                                                                                    if address2 is FAILURE:
                                                                                                                                        self._offset = index2
            if address2 is not FAILURE:
                elements0.append(address2)
                address3 = FAILURE
                address3 = self._read___()
                if address3 is not FAILURE:
                    elements0.append(address3)
                else:
                    elements0 = None
                    self._offset = index1
            else:
                elements0 = None
                self._offset = index1
        else:
            elements0 = None
            self._offset = index1
        if elements0 is None:
            address0 = FAILURE
        else:
            address0 = self._actions.make_opt(self._input, index1, self._offset, elements0)
            self._offset = self._offset
        self._cache['opt_'][index0] = (address0, self._offset)
        return address0

    def _read___(self):
        address0, index0 = FAILURE, self._offset
        cached = self._cache['__'].get(index0)
        if cached:
            self._offset = cached[1]
            return cached[0]
        remaining0, index1, elements0, address1 = 0, self._offset, [], True
        while address1 is not FAILURE:
            chunk0 = None
            if self._offset < self._input_size:
                chunk0 = self._input[self._offset:self._offset + 1]
            if chunk0 is not None and Grammar.REGEX_13.search(chunk0):
                address1 = TreeNode(self._input[self._offset:self._offset + 1], self._offset)
                self._offset = self._offset + 1
            else:
                address1 = FAILURE
                if self._offset > self._failure:
                    self._failure = self._offset
                    self._expected = []
                if self._offset == self._failure:
                    self._expected.append('[\\s]')
            if address1 is not FAILURE:
                elements0.append(address1)
                remaining0 -= 1
        if remaining0 <= 0:
            address0 = TreeNode(self._input[index1:self._offset], index1, elements0)
            self._offset = self._offset
        else:
            address0 = FAILURE
        self._cache['__'][index0] = (address0, self._offset)
        return address0


class Parser(Grammar):
    def __init__(self, input, actions, types):
        self._input = input
        self._input_size = len(input)
        self._actions = actions
        self._types = types
        self._offset = 0
        self._cache = defaultdict(dict)
        self._failure = 0
        self._expected = []

    def parse(self):
        tree = self._read_comp_expression()
        if tree is not FAILURE and self._offset == self._input_size:
            return tree
        if not self._expected:
            self._failure = self._offset
            self._expected.append('<EOF>')
        raise ParseError(format_error(self._input, self._failure, self._expected))


def format_error(input, offset, expected):
    lines, line_no, position = input.split('\n'), 0, 0
    while position <= offset:
        position += len(lines[line_no]) + 1
        line_no += 1
    message, line = 'Line ' + str(line_no) + ': expected ' + ', '.join(expected) + '\n', lines[line_no - 1]
    message += line + '\n'
    position -= len(line) + 1
    message += ' ' * (offset - position)
    return message + '^'

def parse(input, actions=None, types=None):
    parser = Parser(input, actions, types)
    return parser.parse()



import sys
import platform
import inspect
import importlib

import six

from rook.logger import logger
from rook.processor.namespaces.frame_namespace import FrameNamespace
from ..exceptions import RookBdbFailedException


class BdbLocationService(object):
    NAME = "Bdb"

    def __init__(self):
        self._bdb = self._get_bdb().Bdb()
        self._bdb.user_line = user_line

        self._breakpoints = dict()

        self._bdb.set_trace()

    def ignore_current_thread(self):
        self._bdb.ignore_current_thread()

    def add_breakpoint_aug(self, item, lineno, filepath, aug):
        """
        :param item: A module or a code object to add a BP to.
        """
        logger.info("Setting breakpoint (aug ID %s) at %s:%s (%s, object id %u)", aug.aug_id, filepath, lineno,
                    type(item), id(item))

        result = self._bdb.set_break(item, filepath, lineno, aug.aug_id)
        if result:
            raise RookBdbFailedException(result)
        self.mark_aug_active(aug)

    def mark_aug_active(self, aug):
        self._breakpoints[aug.aug_id] = aug
        aug.set_active()

    def remove_aug(self, aug_id):
        # Get current augs in position
        if aug_id not in self._breakpoints:
            return

        self._bdb.clear_break(aug_id)

        aug = self._breakpoints.pop(aug_id)

        # Update status
        aug.set_removed()

    def clear_augs(self):
        # Attempt to clean nicely

        for aug_id in list(self._breakpoints.keys()):
            self.remove_aug(aug_id)

        # Clean hard any left overs
        self._breakpoints = dict()
        self._bdb.clear_all_breaks()

    def close(self):
        self.clear_augs()
        self._bdb.close()

    def pre_fork(self):
        pass

    def post_fork(self):
        pass

    def get_canonical_path(self, module):
        return self._bdb.canonic(inspect.getsourcefile(module))

    @staticmethod
    def _get_bdb():
        from rook.config import InstrumentationConfig
        if InstrumentationConfig.ENGINE == "auto":
            if 'CPython' == platform.python_implementation() and sys.platform in ('darwin', 'linux2', 'linux'):
                if (six.PY2 and sys.version_info > (2, 7, 0)) or (six.PY3 and sys.version_info >= (3, 5, 0)):
                    from . import google_bdb as bdb
                else:
                    from . import py_bdb as bdb
            else:
                from . import py_bdb as bdb
        else:
            bdb = importlib.import_module(".." + InstrumentationConfig.ENGINE, __name__)

        return bdb


# This function has been moved outside of the class so that it can be pickled
# safely by cloudpickle (which will pickle any objects referred to by its closure)
# When changing it, take care to avoid using references to anything not imported within the function
def user_line(frame, filename, aug_id, lineno=None):
    from rook.logger import logger as user_line_logger
    try:
        from rook.singleton import singleton_obj as singleton
        self = singleton.get_trigger_services().get_service(BdbLocationService.NAME)
        # Some Bdb implementations report line numbers before the break
        frame_namespace = FrameNamespace(frame, lineno)

        effective_lineno = frame_namespace._lineno

        aug = self._breakpoints.get(aug_id)

        if aug:
            aug.execute(frame_namespace, dict())
        else:
            user_line_logger.error("Aug not found! %s@%d (aug id %r)", str(filename), effective_lineno, aug_id)
    except:  # lgtm[py/catch-base-exception]
        user_line_logger.exception("Error while processing breakpoint")


import sys
import six
import traceback


def rook_test():
    six.print_("[Rookout] Testing connection to agent")

    if '-v' in sys.argv:
        from rook.config import LoggingConfiguration

        LoggingConfiguration.LOG_LEVEL = "DEBUG"
        LoggingConfiguration.LOG_TO_STDERR = True

    from rook.config import AgentComConfiguration
    AgentComConfiguration.TIMEOUT = 10

    try:
        import rook

        rook.start(throw_errors=True)
        success = True
    except:  # lgtm[py/catch-base-exception]
        six.print_(file=sys.stderr)
        traceback.print_exc()
        six.print_(file=sys.stderr)

        success = False

    if success:
        six.print_("[Rookout] Test Finished Successfully")
        exit(0)
    else:
        six.print_("[Rookout] Test Failed")
        exit(1)


if '__main__' == __name__:
    rook_test()



from .namespace import Namespace
from .container_namespace import ContainerNamespace
from rook.logger import logger
from .python_object_namespace import PythonObjectNamespace

SPAN_ATTRIBUTES = [ 'tags',
                    'metrics',
                    'service',
                    'name',
                    'resource',
                    'span_id',
                    'trace_id',
                    'parent_id',
                    '_parent',
                    'meta',
                    'error',
                    'metrics',
                    'span_type',
                    'start_ns',
                    'duration_ns',
                    'duration',
                    'operation_name',
                    'finished',
                    'context',
                    '_context',
                    'start_ns',
                    'start_time',
                    'end_time',
                    'type']


class TraceNamespace(Namespace):
    def __init__(self):
        super(TraceNamespace, self).__init__(self.METHODS)
        self.span = None

    def dump(self, args):
        if not self._load():
            return PythonObjectNamespace("Could not fetch span")

        # Check if its jaeger span since jaeger span implement variant like tags
        if self.is_jaeger_span(self.span):
            return self.dump_jaeger_span(self.span)

        return self.dump_generic_span(self.span)

    def read_attribute(self, name):
        if not self._load():
            return PythonObjectNamespace("Could not fetch span")

        if name == 'span':
            return PythonObjectNamespace(self.span)

        if hasattr(self.span, name):
            return PythonObjectNamespace(getattr(self.span, name))

        # TODO::FUTURE throw the exception
        # raise RookAttributeNotFound(name)
        logger.debug("Attribute Not Found %s", name)
        return PythonObjectNamespace(None)

    def _load(self):
        if self.span is not None:
            return True

        try_generic_opentrace = True
        try:
            # Try to fetch ddtrace first since if this will work - opentracing.tracer.active_span will work as well
            # but won't return the right span
            import ddtrace
            self.span = ddtrace.tracer.get_call_context().get_current_span()

            if self.span:
                try_generic_opentrace = False

        except ImportError:
            pass # We'll try the generic opentrace next

        if try_generic_opentrace:
            try:
                import opentracing
                self.span = opentracing.tracer.active_span
            except ImportError:
                pass

            if not self.span:
                try:
                    from opentelemetry import trace
                    self.span = trace.get_current_span()
                except ImportError:
                    return False

        if self.span:
            return True
        return False

    @staticmethod
    def is_jaeger_span(span):
        try:
            import jaeger_client
            return isinstance(span, jaeger_client.span.Span)
        except Exception:
            pass
        return False

    @staticmethod
    def dump_jaeger_span(span):
        jaeger_tags = {}

        for tag in span.tags:
            # I do not filter keys in the sdk - the fe will parse it later.
            key = tag.key
            value = None
            if tag.vType == 0:
                value = tag.vStr
            elif tag.vType == 1:
                value = tag.vDouble
            elif tag.vType == 2:
                value = tag.vBool
            elif tag.vType == 3:
                value = tag.vLong
            elif tag.vType == 4:
                value = tag.vBinary
            jaeger_tags[key] = value

        return ContainerNamespace(
            {
                "tags": PythonObjectNamespace(jaeger_tags),
                "span_id": PythonObjectNamespace(span.span_id),
                "parent_id": PythonObjectNamespace(span.parent_id),
                "operation_name": PythonObjectNamespace(span.operation_name),
                "finished": PythonObjectNamespace(span.finished),
                "start_time": PythonObjectNamespace(span.start_time),
                "end_time": PythonObjectNamespace(span.end_time),
                "context": PythonObjectNamespace(span.context),
            }
        )

    @staticmethod
    def get_span_attribute(span, attr):
        if hasattr(span, attr):
            return PythonObjectNamespace(getattr(span, attr),
                                         PythonObjectNamespace.ObjectDumpConfig.tolerant_limits(None))

        return PythonObjectNamespace(None)

    @staticmethod
    def dump_generic_span(span):
        attributes = {}
        for attribute in SPAN_ATTRIBUTES:
            attributes[attribute] = TraceNamespace.get_span_attribute(span, attribute)

        return ContainerNamespace(attributes)

    METHODS = (dump, )



import threading
import time
from six.moves.urllib.parse import urlparse
import certifi
import ssl
import re
import os
import six
import websocket

from rook.com_ws.flush_messages_event import FlushMessagesEvent
from rook.com_ws.singlethread.network_event_loop import NetworkEventLoop
from rook.exceptions import RookCommunicationException, RookInvalidToken, RookDependencyConflict, RookMessageSizeExceeded, RookMissingToken
import rook.com_ws.socketpair_compat  # do not remove - adds socket.socketpair on Windows lgtm[py/unused-import]


try:
    from websocket import WebSocketBadStatusException # This is used to make sure we have the right version lgtm[py/unused-import]
except ImportError:
    raise RookDependencyConflict('websocket')

# Python < 2.7.9 is missing important SSL features for websocket
# (unless supplied by CentOS etc)
if not websocket._ssl_compat.HAVE_SSL:
    try:
        import backports.ssl
        import backports.ssl_match_hostname
        websocket._http.ssl = backports.ssl
        websocket._http.HAVE_SSL = True
        websocket._http.HAVE_CONTEXT_CHECK_HOSTNAME = True
    except ImportError:
        six.print_('[Rookout] Python is missing modern SSL features. To rectify, please run:\n'
                   '  pip install rook[ssl_backport]')

from rook.com_ws import information, selectable_event, selectable_queue
from rook.logger import logger
import rook.protobuf.messages_pb2 as messages_pb
import rook.protobuf.envelope_pb2 as envelope_pb
from rook.config import AgentComConfiguration, VersionConfiguration


def wrap_in_envelope(message):
    envelope = envelope_pb.Envelope()
    envelope.timestamp.GetCurrentTime()
    envelope.msg.Pack(message)

    return envelope.SerializeToString()


class MessageCallback(object):
    def __init__(self, cb, persistent):
        self.cb = cb
        self.persistent = persistent


class AgentCom(object):
    def __init__(self, agent_id, host, port, proxy, token, labels, tags, debug, print_on_initial_connection):
        self.id = agent_id

        self._host = host if '://' in host else 'ws://' + host
        self._port = port
        self._proxy = proxy
        self._token = token
        self._token_valid = False
        self._labels = labels or {}
        self._tags = tags or []

        self._loop = None
        self._connection = None
        self._queue = selectable_queue.SelectableQueue()

        self._running = True

        self._ready_event = threading.Event()
        self._connection_error = None

        self.debug = debug

        self._callbacks = {}

        self._thread = None
        self._print_on_initial_connection = print_on_initial_connection

        def set_ready_event(*args):
            self._ready_event.set()

        self.once('InitialAugsCommand', set_ready_event)

    def start(self):
        self._init_connect_thread()
        self._thread.start()

    def stop(self):
        self._running = False

        if self._connection is not None:
            self._connection.shutdown()

        if self._loop is not None:
            loop = self._loop
            self._loop = None
            loop.stop()

        if self._thread is not None:
            self._thread.join()
            self._thread = None

    def restart(self):
        self.stop()

        self._connection = None

        self._running = True
        self.start()

    def update_info(self, agent_id, tags, labels):
        self.id = agent_id
        self._labels = labels or {}
        self._tags = tags or []

    def add(self, message):
        if self._queue.qsize() >= AgentComConfiguration.MAX_QUEUED_MESSAGES:
            return None

        packaged_message = wrap_in_envelope(message)
        if len(packaged_message) > AgentComConfiguration.MAX_MESSAGE_LENGTH:
            return RookMessageSizeExceeded(len(packaged_message), AgentComConfiguration.MAX_MESSAGE_LENGTH)

        self._queue.put(packaged_message)
        return None

    def is_queue_full(self):
        return self._queue.qsize() >= AgentComConfiguration.MAX_QUEUED_MESSAGES

    def on(self, message_name, callback):
        self._register_callback(message_name, MessageCallback(callback, True))

    def once(self, message_name, callback):
        self._register_callback(message_name, MessageCallback(callback, False))

    def await_message(self, message_name):
        event = selectable_event.SelectableEvent()
        self.once(message_name, lambda _: event.set())

        return event

    def wait_for_ready(self, timeout=None):
        if not self._ready_event.wait(timeout):
            raise RookCommunicationException()
        else:
            if self._connection_error is not None:
                raise self._connection_error

    def _network_loop(self):
        retry = 0
        backoff = AgentComConfiguration.BACK_OFF
        connected = False
        last_successful_connection = 0

        while self._running:
            try:
                try:
                    if connected and time.time() >= last_successful_connection + AgentComConfiguration.RESET_BACKOFF_TIMEOUT:
                        retry = 0
                        backoff = AgentComConfiguration.BACK_OFF

                    self._connection = self._create_connection()

                    self._register_agent(self.debug)

                except websocket.WebSocketBadStatusException as e:
                    if not self._token_valid and e.status_code == 403:# invalid token
                        if self._token is None:
                            self._connection_error = RookMissingToken()
                        else:
                            self._connection_error = RookInvalidToken(self._token)
                        self._ready_event.set()

                        logger.error('Connection failed; %s', self._connection_error.get_message())
                    raise
            except Exception as e:
                retry += 1
                backoff = min(backoff * 2, AgentComConfiguration.MAX_SLEEP)
                connected = False

                if hasattr(e, 'message') and e.message:
                    reason = e.message
                else:
                    reason = str(e)

                logger.info('Connection failed; reason = %s, retry = #%d, waiting %.3fs', reason, retry, backoff)

                time.sleep(backoff)
                continue
            else:
                connected = True
                last_successful_connection = time.time()

            if self._print_on_initial_connection:
                # So there is no print on reconnect
                self._print_on_initial_connection = False
                six.print_("[Rookout] Successfully connected to controller.")
            logger.debug("WebSocket connected successfully")
            self._token_valid = True
            with self.await_message('InitialAugsCommand') as got_initial_augs_event:
                if self._loop is not None:
                    self._loop.stop()
                    self._loop = None
                self._loop = NetworkEventLoop(self._connection, self._queue, self._handle_incoming_message, got_initial_augs_event)
                self._loop.run_until_stopped()

            if self._running:
                logger.debug("Reconnecting")

    def flush_all_messages(self):
        flush_event = FlushMessagesEvent()
        self._queue.put(flush_event)
        flush_event.event.wait(AgentComConfiguration.FLUSH_TIMEOUT)

    def _create_connection(self):
        url = '{}:{}/v1'.format(self._host, self._port)
        headers = {
            'User-Agent': 'RookoutAgent/{}+{}'.format(VersionConfiguration.VERSION, VersionConfiguration.COMMIT)
        }

        if self._token is not None:
            headers["X-Rookout-Token"] = self._token

        proxy_host, proxy_port = self._get_proxy()

        connect_args = (url,)
        connect_kwargs = dict(header=headers,
                              timeout=AgentComConfiguration.TIMEOUT,
                              http_proxy_host=proxy_host, http_proxy_port=proxy_port)

        if os.environ.get('ROOKOUT_NO_HOST_HEADER_PORT') == '1':
            host = re.sub(':\d+$', '', urlparse(url).netloc)
        else:
            host = None

        connect_kwargs['sslopt'] = dict()

        if os.environ.get('ROOKOUT_SKIP_SSL_VERIFY') == '1':
            connect_kwargs['sslopt']['cert_reqs']=ssl.CERT_NONE

        try:
            # connect using system certificates
            conn = websocket.create_connection(*connect_args, host=host, **connect_kwargs)
        # In some very specific scenario, you cannot
        # reference ssl.CertificateError because it does
        # exist, so instead we get with with getattr
        # (None never matches an exception)
        except (ssl.SSLError, getattr(ssl, 'CertificateError', None)):
            # connect using certifi certificate bundle
            # (Python 2.7.15+ from python.org on macOS rejects our CA, see RK-3383)
            connect_kwargs['sslopt']['ca_certs']=certifi.where()
            logger.debug("Got SSL error when connecting using system CA cert store, falling back to certifi")
            conn = websocket.create_connection(*connect_args, **connect_kwargs)
        # just in case there's a bug and we get stuck - timeout should never be hit since we always use select()
        conn.settimeout(30)
        return conn

    def _get_proxy(self):
        if self._proxy is None:
            return None, None

        try:
            if not self._proxy.startswith("http://"):
                self._proxy = "http://" + self._proxy

            url = urlparse(self._proxy, "http://")

            logger.debug("Connecting via proxy: %s", url.netloc)

            return url.hostname, url.port
        except ValueError:
            return None, None

    def _register_agent(self, debug):
        logger.info('Registering agent with id %s', self.id)
        info = information.collect(debug)
        info.agent_id = self.id
        info.labels = self._labels
        info.tags = self._tags
        
        m = messages_pb.NewAgentMessage()
        m.agent_info.CopyFrom(information.pack_agent_info(info))

        return self._send(wrap_in_envelope(m))

    def _init_connect_thread(self):
        if self._thread is not None:
            raise RuntimeError('Trying to start AgentCom thread twice')

        self._thread = threading.Thread(name="rookout-" + type(self).__name__, target=self._network_loop)
        self._thread.daemon = True

    AcceptedMessageTypes = [
        messages_pb.InitialAugsCommand,
        messages_pb.AddAugCommand,
        messages_pb.ClearAugsCommand,
        messages_pb.PingMessage,
        messages_pb.RemoveAugCommand
    ]

    def _handle_incoming_message(self, envelope):
        for message_type in self.AcceptedMessageTypes:
            if envelope.msg.Is(message_type.DESCRIPTOR):
                message = message_type()
                envelope.msg.Unpack(message)
                type_name = message.DESCRIPTOR.name

                callbacks = self._callbacks.get(type_name)

                if callbacks:
                    persistent_callbacks = []

                    # Trigger all persistent callbacks first
                    for callback in callbacks:
                        try:
                            if callback.persistent:
                                callback.cb(message)
                        except Exception:  # We ignore errors here, they are high unlikely and the code is too deep
                            pass
                        finally:
                            if callback.persistent:
                                persistent_callbacks.append(callback)

                    # Trigger all non persistent callbacks
                    for callback in callbacks:
                        try:
                            if not callback.persistent:
                                callback.cb(message)
                        except Exception:  # We ignore errors here, they are high unlikely and the code is too deep
                            pass

                    self._callbacks[type_name] = persistent_callbacks

    def _send(self, message):
        return self._connection.send_binary(message)

    def _register_callback(self, message_name, callback):
        self._callbacks.setdefault(message_name, []).append(callback)



from rook.exceptions import RookWriteAttributeNotSupported, RookAttributeNotFound, RookMethodNotFound


class Namespace(object):
    def __init__(self, methods=None):
        self._methods = {}

        if methods:
            for method in methods:
                self._methods[method.__name__] = method

    def call_method(self, name, args):
        try:
            method = self._methods[name]
        except KeyError:
            raise RookMethodNotFound(type(self), name)

        return method(self, args)

    def read_attribute(self, name):
        try:
            if hasattr(self, '__dict__') and self.__dict__ and name in self.__dict__:
                return self.__dict__[name]
            if hasattr(self, '__slots__') and self.__slots__ and name in self.__slots__:
                return self.__slots__[name]

            raise RookAttributeNotFound(name)
        except AttributeError:
            raise RookAttributeNotFound(name)

    def write_attribute(self, name, value):
        raise RookWriteAttributeNotSupported(type(self), name)

    def read_key(self, key):
        try:
            return self[key]
        except KeyError:
            raise RookAttributeNotFound(key)



import sys
import time

def seconds_to_int_nano_seconds(seconds):
    '''
    Used to convert time.time() from seconds to nanoseconds.
    Note that Python>=3.7 has time.time_ns() that returns exactly the time in nanoseconds. We use this function for
    backwards compatibility.
    '''
    return int((10**9)*seconds)

def milliseconds_to_int_nano_seconds(milliseconds):
    return int((10**6)*milliseconds)

'''
This trick here allows to check the python version only once when the module is loaded,
and call get_most_accurate_time_stamp_nano_seconds() to find what is the time stamp.
'''
if sys.version_info>=(3,7):
    get_most_accurate_time_stamp_nano_seconds = lambda: time.perf_counter_ns()
elif sys.version_info>=(3,3):
    get_most_accurate_time_stamp_nano_seconds = lambda: seconds_to_int_nano_seconds(time.perf_counter())
else:
    get_most_accurate_time_stamp_nano_seconds = lambda: seconds_to_int_nano_seconds(time.time())



from .namespace import Namespace
from .python_object_namespace import PythonObjectNamespace
from rook.protobuf import variant_pb2


class TracebackNamespace(Namespace):
    def __init__(self, frame, depth):
        super(TracebackNamespace, self).__init__()
        self._frame = frame
        self._depth = depth

    def __getitem__(self, key):
        return self.read_key(key)

    def call_method(self, name, args):
        if name == "size":
            return PythonObjectNamespace(self._depth)
        else:
            return super(TracebackNamespace, self).call_method(name, args)

    def read_key(self, key):
        pos = int(key)

        current_frame = self._frame

        for i in range(pos):
            current_frame = current_frame.f_back()

        return current_frame

    def dump(self, variant, getStringIndexInCache=None):
        variant.variant_type = variant_pb2.Variant.Type.VARIANT_TRACEBACK

        current_frame = self._frame

        for i in range(self._depth):
            frame = variant_pb2.Variant.CodeObject()

            frame.lineno = current_frame.line().obj if current_frame.line().obj is not None else 0
            if getStringIndexInCache is None:
                frame.filename = current_frame.filename().obj if current_frame.filename().obj is not None else "unavailable"
                frame.name = current_frame.function().obj if current_frame.function().obj is not None else "unavailable"
                frame.module = current_frame.module().obj if current_frame.module().obj is not None else "unavailable"
            else:
                frame.filename_index_in_cache = getStringIndexInCache(current_frame.filename().obj if current_frame.filename().obj is not None else "unavailable")
                frame.name_index_in_cache = getStringIndexInCache(current_frame.function().obj if current_frame.function().obj is not None else "unavailable")
                frame.module_index_in_cache = getStringIndexInCache(current_frame.module().obj if current_frame.module().obj is not None else "unavailable")

            variant.traceback.locations.append(frame)

            current_frame = current_frame.f_back()

            if not current_frame:
                break



"""This module initializes the flask hook """

import six
import time

from rook.logger import logger

from rook.processor.namespaces.container_namespace import ContainerNamespace
from rook.processor.namespaces.python_object_namespace import PythonObjectNamespace

from functools import update_wrapper


class FlaskTracer(object):
    NAME = "flask"

    def __init__(self, execute_augs_callback):
        logger.info("installing flask hook")

        try:
            import flask
            from flask import request
            self._request = request
        except ImportError as e:
            six.print_("Rook failed to import dependencies: " + str(e))
            return

        if getattr(flask, '_flask_patch', False):
            return

        self._execute_augs_callback = execute_augs_callback

        # Install hooks
        flask.app.Flask.__init__ = self._wrap_function(self, flask.app.Flask.__init__)

        setattr(flask, '_flask_patch', True)

    @staticmethod
    def _wrap_function(flask_tracer, function_to_wrap):
        # self is the app instance
        def new_f(self, import_name, *args, **kwargs):

            ret_val = function_to_wrap(self, import_name, *args, **kwargs)

            self.before_request(flask_tracer.before_request)
            self.after_request(flask_tracer.after_request)
            self.teardown_request(flask_tracer.teardown_request)

            return ret_val

        return update_wrapper(new_f, function_to_wrap)

    def before_request(self):
        try:
            setattr(self._request, "start_time", time.time())
        except:  # lgtm[py/catch-base-exception]
            return

    def after_request(self, response):
        try:
            code = response.status_code if response else ''
            setattr(self._request, "status_code", code)
        except Exception: # Ignore errors here fro now
            pass

        return response

    def teardown_request(self, exception):
        try:
            start_time = getattr(self._request, 'start_time', None)
            if not start_time:
                return

            code = getattr(self._request, 'status_code', None)
            if not code:
                code = "Unknown"

            duration = time.time() - start_time
            extracted = ContainerNamespace(
                {'duration': PythonObjectNamespace(duration),
                 'remote_host': PythonObjectNamespace(self._request.remote_addr),
                 'url': PythonObjectNamespace(self._request.url),
                 'method': PythonObjectNamespace(self._request.method),
                 'status_code': PythonObjectNamespace(code)})

            self._execute_augs_callback(extracted)
        except:  # lgtm[py/catch-base-exception]
            return
